import copy
from structures import Tree, Dag
import graphutils
import pprint
import pydot
import math
import itertools
from Queue import PriorityQueue
from collections import deque
from dictutils import frozendict



from pyparsing import Word, alphas, nums, alphanums, Suppress, Literal, delimitedList, Forward, Optional, Empty, FollowedBy, ZeroOrMore, Group

VERBOSE = False
#VERBOSE = True


def treeify(l):
    from parse import tree_from_string
    if len(l) == 2:
        return tuple(l)
    elif len(l) == 1:
        return (l[0], tree_from_string("0"))
    else:
        raise ValueError("something wrong with this rule")

class Rule:

    """Currently not much more than a dictionary"""
    def __init__(self, heads, pattern, tails, weight=1.0):
#        print heads, pattern, tails
        self.weight = - math.log(float(weight))
        self.pattern = "%s:%s" % (pattern, ":".join(t[0] for t in tails))
        self.heads = tuple(treeify(h) for h in heads) # state-tree pairs
        self.tails = tuple(t[1][0] for t in tails) # state-tree pairs
    
    def __str__(self):
        psplit = self.pattern.split(":")
        label = psplit[0]
        roles = psplit[1:]
        rolestails = " ".join(":%s %s" % (r, s) for (r, s) in zip(roles, self.tails))
#        print rolestails, self.heads
        return "%s -> (%s %s) @ %s" % (
             " ".join("%s.%s" % (h, t) for (h, t) in self.heads), 
             label, rolestails, math.e ** (- self.weight))
            
            
class BDT:

    def __and__(self, other):
        return self.intersection(other)

    def intersection(self, other):
        from parse import tree_from_string
        self_rules = self.rules
        other_rules = other.rules
        new_rules = []
        for r1 in self_rules:
            for r2 in other_rules:
                if (r1.pattern == r2.pattern and
                    len(r1.heads) == len(r2.heads) and
                    len(r1.tails) == len(r2.tails)):
                    psplit = r1.pattern.split(':')
                    label = psplit[0]
                    roles = psplit[1:]
#                    print label, roles
                    # can synchronize
                    new_heads = [tuple(zip(a,b)) 
                        for (a,b) in zip(r1.heads, r2.heads)]
                    new_heads = tuple(("(%s,%s)" % (a,b),) 
                        for ((a,b),(c,d)) in new_heads)
                    new_tails = tuple((a, ("(%s,%s)" % (b,c),))
                        for (a,b,c) in zip(roles, r1.tails, r2.tails))
                    w = math.e ** (- r1.weight - r2.weight)
                    new_rules.append(Rule(new_heads, label, new_tails, w))
        new_automaton = BDT(rule_list = new_rules, weighted = True)
        new_automaton.initial_state = "(%s,%s)" % (
            self.initial_state, other.initial_state)
        return new_automaton

    def __or__(self, other):
        return self.union(other)
        
    def union(self, other):
        new_rules = []
        new_initial_state = "s"
        for (i, automaton) in enumerate((self, other)):
            for rule in automaton.rules:
#                print rule.pattern
                psplit = rule.pattern.split(':')
                label = psplit[0]
                roles = psplit[1:]
                new_tails = tuple((role, ("%s%s" % (tail, i),))
                    for (role, tail) in zip(roles, rule.tails))
                new_heads = tuple(("%s%s" % (a, i),) for (a,b) in rule.heads)
                new_rules.append(Rule(new_heads, label,
                    new_tails, math.e ** (- rule.weight)))
                if tuple(a for (a,b) in rule.heads) == (automaton.initial_state,):
                    new_rules.append(Rule((new_initial_state,),
                        label, new_tails, math.e ** (- rule.weight)))
        new_automaton = BDT(rule_list = new_rules)
        new_automaton.initial_state = new_initial_state
        return new_automaton

    def backward(self, number, tree):
        return self.generate(number=number, backward=tree)

    def generate(self, state = None, number = 1, genfile = None, random = False, test = None, backward = None):
        if state == None: state = self.initial_state
        dag_counter = 0
        counter = 0
        graphs = []
        initial_dag = Dag()
        v = initial_dag.add_vertex(None, 'root')
        initial_dag.V[v].tails = [state]
        port = (v, 1) # positive number indicates tail
        # use number of ports to keep priority queue sorted
        initial_c = dict(dag_id = dag_counter, ports = [port],
            dag = initial_dag, tree = Tree(port, []),
            tree_map = dict([(port, Tree(port, []))]))
        q = PriorityQueue()
        q.put((0, initial_c))
        qsize = 1
        while counter < number:
            try:
                (w, c) = q.get()
                (weight, dag_id, ports, dag, tree, tree_map) = (w, 
                    c['dag_id'], c['ports'], c['dag'], c['tree'], c['tree_map'])
                if VERBOSE:
                    print "graph %s taken from queue" % dag_id
            except:
                print "empty queue"
                return None
            

            qsize -= 1  

            if ports: # incomplete graph needs expanding

                for (rule_id, rule) in enumerate(self.rules):
                    #print rule, rule.weight #, dag
                    heads = [h for (h, t) in rule.heads]
                    tails = list(rule.tails)
                    vports = tails + heads
                    #print "rule: %s" % vports, ports
                    for subset in itertools.product(*[range(len(ports) + 1)
                      for _ in range(len(vports))]):
                        #print subset, 
                        # we've now got for each head and tail of the
                        # new vertex exactly one port or 0
                        sublist = [i for i in subset if i > 0]

                        if (len(sublist) == len(set(sublist)) 
                            and (1 in sublist)):
                          if (all((subset[i] == 0 
                               # 0 means: add a new open port
                               or (i < len(tails) and
                               (ports[subset[i] - 1][1] < 0 and
                               # we want to connect a tail to a head
                               dag.V[ports[subset[i] - 1][0]].
                               heads[(- ports[subset[i] - 1][1]) - 1] == vports[i]))
                               or (i >= len(tails) and 
                               (ports[subset[i] - 1][1] > 0 and
                               # we want to connect a head to a tail
                               dag.V[ports[subset[i] - 1][0]].
                               tails[ports[subset[i] - 1][1] - 1] == vports[i])))
                                  for i in range(len(subset)))):

                            # rebuild rule tree
                            mapping = {}
                            new_ports = [port for (i, port)
                              in enumerate(ports) if (i + 1) not in subset]
                              # keep unused ports
                            new_dag = copy.deepcopy(dag)
                            new_v = new_dag.add_vertex(None, rule.pattern)
                            # new vertex has to be connected to existing ports
                            # or have new open ports
                            new_dag.V[new_v].tails = tails[:]
                            new_dag.V[new_v].heads = heads[:]

                            new_tree_map = copy.deepcopy(tree_map)

                            #print vports, subset, ports
                            for i in range(len(subset)):                        
                                if i < len(tails):
                                    if subset[i] == 0:
                                        # add new tail open port
                                        new_port = (new_v, i + 1)
                                        new_ports.append(new_port)
                                        #print tree_map, mapping, tree, 1
                                        mapping[str(i + 1)] = Tree(new_port, [])
                                    else:#
                                        # connect tail to existing head open port
                                        port, index = ports[subset[i] - 1]
                                        new_e = new_dag.add_edge(new_v, port, 
                                      existing = True)
                                        new_dag.V[new_v].tails[i] = new_e
                                        new_dag.V[port].heads[(- index) - 1] = new_e
                                        mapping[str(i + 1)] = tree_map[(port, index)]
                                else:
                                    if subset[i] == 0:
                                        # add new head open port
                                        new_port = (new_v, -(i - len(tails) + 1))
                                        new_ports.append(new_port)
                                        new_tree_map[new_port] = rule.heads[- (i - len(tails))][1] # side-condition on the new head port
                                    else:
                                        # connect head to existing tail open port
                                        port, index = ports[subset[i] - 1]
                                        new_e = new_dag.add_edge(port, new_v, 
                                      existing = True)
                                        new_dag.V[new_v].heads[-(i - len(tails))] = new_e
                                        new_dag.V[port].tails[index - 1] = new_e
                                        new_tree_map[(port, index)] = rule.heads[- (i - len(tails))][1]
                            
                            for (k, v) in new_tree_map.iteritems():
                                new_tree_map[k] = v.substitute(mapping)
                            new_tree = tree.substitute(new_tree_map)
                            
                            if (not backward or
                                backward.compatible(new_tree,
                                (lambda x: x in new_ports))):
                                try:
                                    new_dag.topsort_vertices()
                                    dag_counter += 1

                                    if self.is_weighted:
                                        nweight = weight + rule.weight
                                    else:
                                        nweight = len(new_ports)
                                    q.put((nweight, dict(dag_id = dag_counter, ports = new_ports, dag = new_dag, tree = new_tree, tree_map = new_tree_map)))
                                    qsize += 1
                                except(graphutils.CycleError):
                                    break

                                if VERBOSE:
                                    print "new graph %03d with weight %s put on queue" % (dag_counter, math.e ** (-nweight)), new_dag
                            #new_dag.dot().write_pdf("debug/incomplete_dag%03d.pdf" % dag_counter)
            else:
                counter += 1
                if self.is_weighted:
                    nweight = math.e ** (-weight)
                else:
                    nweight = "n/a"
                if VERBOSE:
                    print "complete graph #%s, weight %s, queue size %s, incomplete graphs %s" % (counter, nweight, qsize, dag_counter)
                #print tree
                #print dag.V, dag.E
                if test:
                    try:
                        parse = test.evaluate(dag)
                        results = parse['results']
                        result = list(t.frontier() for (s, t) in results)
                    except(graphutils.CycleError):
                        print "bad graph"
                        result = []
                    
                if VERBOSE:
                    print dag.penman()    
                dag.remove_pseudoroot()
                graphs.append(dag)
                if VERBOSE: print "#%03d %s" % (len(graphs), dag.penman())
                if test: print result
                #dag.dot().write_pdf("pdf/complete_dag%05d.pdf" % counter)
                if genfile:
                    (prefix, suffix) = genfile.split('.', 1)
                    dag.to_file('%s.%05d.%s' % 
                      (prefix, counter, suffix))
                if counter == number: return graphs
                    
                         
        return graphs

    def connect(self, identicals):
        # joins identical states
        id_dict = {}
        eq_set = {}
        # map every state onto representative     
        for (p, q) in identicals:
            if p not in eq_set:
                eq_set[p] = set([p])
            if q not in eq_set:
                eq_set[q] = set([q])
                     
            eq_set[p].update(eq_set[q])
            for state in eq_set[p]:
                id_dict[state] = p
                eq_set[state] = eq_set[p] 

        for rule in self.rules:
            rule.tails = tuple(id_dict[p] if p in id_dict else p for p in rule.tails)
            rule.heads = tuple((id_dict[p], t) if p in id_dict else (p, t) for (p, t) in rule.heads)
        self.sort_ruleset()
        
    def dot(self, with_labels=False, bottomup=True):
        return self.as_graph().dot(with_labels, bottomup)

    def as_graph(self):
        graph = Dag()
        vertexlookup = {} # don't add vertices twice
    
        for (i, rule) in enumerate(self.rules):
            v = graph.add_vertex("%s.%s" % (i, rule.pattern), vlabel=rule.pattern)
            for (ii, h) in enumerate(rule.heads):
                h = h[0]
                if h not in vertexlookup:
                    vertexlookup[h] = graph.add_vertex(h, vlabel=h)
                graph.add_edge(v, vertexlookup[h], label=str(ii+1))
            for (ii, t) in enumerate(rule.tails):
                if t not in vertexlookup:
                    vertexlookup[t] = graph.add_vertex(t, vlabel=t)
                graph.add_edge(vertexlookup[t], v, label=str(ii+1))
        return graph
                
    def sort_ruleset(self):
        self.ruleset = {}
        for (i, rule) in enumerate(self.rules):
            lhs = (rule.pattern, rule.tails)
            if lhs not in self.ruleset:
                 self.ruleset[lhs] = set()
            self.ruleset[lhs].add(i)

    def __init__(self, rule_list = None, from_file = None, transducer = False, weighted = False, initial_state=None):
        self.is_weighted = weighted
        self.rules = []
        self.initial_state = initial_state
        if rule_list:
            self.rules.extend(rule_list)
        self.sort_ruleset()
        
    def __str__(self):
        return "%s\n%s" % (self.initial_state, 
                         "\n".join(str(r) for r in self.rules))
    
    def __repr__(self):
        return self.__str__()
        
       
    def evaluate(self, dag):
      try:
        # helper functions
        def mother_c(c, v):
            return c if v == c['vertex'] else mother_c(c['mother'], v)
           
        def get_tails(c, v):
            return tuple(c['edges'][e][0] for e in dag.V[v].tails)         
            
        def update_c(mother_c, v, heads, weight = 0):            
            mother_v = mother_c['vertex']
            mother_e = mother_c['edges']  
            mother_w = mother_c['weight']          
            new_e = dict(mother_e)
            mapping = {}
            
            for (i, h) in enumerate(dag.V[v].tails):
                # construct the tree mapping
                mapping[str(i + 1)] = mother_e[h][1]
                # remove the states from the head trees
                new_e[h] = (None, None)
            
            for (i, h) in enumerate(dag.V[v].heads):
                # insert the tail trees into rhs
                (state, tree) = heads[i]
                tree = tree.substitute(mapping)
                new_e[h] = (state, tree)
            
            return frozendict([('vertex', v),
                ('edges', frozendict(new_e)), ('mother', mother_c),
                ('weight', mother_w + weight)])
                     
        # The initial configuration
        initial_e = frozendict([(i, (None, None)) for i in range(len(dag.E))])
        initial_c = frozendict([('edges', initial_e),
            ('mother', None), ('vertex', None), ('weight', 0)])

        ccounter = 0
        current_cs = dict([(initial_c, ccounter)])
        cs = {}
        
        deriv_rules = []
        identicals = set()
        rtg_rules = []
        
        # For each vertex in reverse topological order
        for vid in reversed(dag.topsort_vertices()):
            
            v = dag.V[vid]
            vlabel = v.label            
            if vid == dag.root:
                break
            # start a set of new (active for next vertex) configurations
            new_cs = {}
                        
            # For each current configuration                
            for (c, cid) in current_cs.iteritems():
                ctails = get_tails(c, vid)
                if VERBOSE: print vlabel, ctails, self.ruleset.keys()
                if (vlabel, ctails) in self.ruleset:
                    for rule_id in self.ruleset[(vlabel, ctails)]:
                        rule_heads = self.rules[rule_id].heads
                        if len(dag.V[vid].heads) == len(rule_heads):
                            new_c = update_c(c, vid, rule_heads, weight =
                                self.rules[rule_id].weight if self.is_weighted else 0 )
                            # TODO: Merge configurations with same state/tree
                            if new_c not in new_cs:
                                ccounter += 1
                                new_cs[new_c] = new_cid = ccounter
                                new_vtoc = {}
                                    
                            else:
                                if VERBOSE: print "HELP HELP I'M TRAPPED! \n"
                                new_cid = new_cs[new_c]

#                            # rtg rules
#                            mapping = dict((str(i+1),
#                                str(cs[mother_c(c, dag.E[t].tail)])) for
#                                (i, t) in enumerate(v.tails))
#                            
##                            print mapping
#                            for (i, h) in enumerate(rule_heads):
#                                (state, tree) = h
#                                #print state, tree
#                                new_rules, new_identicals = tree.ruleset(
#                                    "c%s" % (str(new_cid),), mapping)
#                                #for r in new_rules:
#                                   #print r.label, r.tails, r.heads
#                                #print new_identicals
#                                rtg_rules.extend(new_rules)
#                                identicals.update(new_identicals)
#                            
                            deriv_rules.append(Rule(tuple(
                                 ("c%s" % (str(new_cid),),) for h in v.heads),
                                 "r%s" % (rule_id,),
                                 tuple(("arg%s" % (i+1,),
                                    ("c%s" % (str(cs[mother_c(c,
                                    dag.E[t].tail)])), 0))
                                    for (i, t) in enumerate(v.tails))
                                 ))
                            

                                
            cs.update(new_cs)
            current_cs = new_cs
            if VERBOSE: pprint.pprint(list((c, hash(c)) for c in new_cs))
        
#        print rtg_rules
        deriv_acceptor = BDT(rule_list = deriv_rules)         
        rtg = BDT(rule_list = rtg_rules)
        rtg.connect(identicals)
        
        unsorted_results = [(c['edges'][t], math.e ** (- c['weight'])) for t in dag.V[dag.root].tails for c in current_cs]
        results = {}
        for ((s, t), w) in unsorted_results:
            if s == self.initial_state:
                if (s, t) in results:
                    results[(s, t)] += w
                else:
                    results[(s, t)] = w
        
        results = [((s, t), str(w)) for ((s, t), w) in results.iteritems()]
        
#        results = [c['edges'][t] for t in dag.V[dag.root].tails for c in current_cs]
            
        
        return {'derivations': deriv_acceptor,
            'results': results, 
            'weighted': self.is_weighted,
            'trees': rtg
            }
      except Exception, e:
        print e

