import graphutils
import pydot

class Tree:

    def ruleset(self, state, mapping):
    # currently broken!
#      try:
        from machines import Rule
#        from parse import tree_from_string
        children_ruleset = []
        identicals = set()
        if self.label in mapping:
            identicals.add((mapping[self.label], state))
        else:
            child_states = []
            for (i, child) in enumerate(self.children):
                child_state = "%s_%s" % (state, i)
                child_states.append(child_state)
                child_ruleset, child_identicals = child.ruleset(
                    child_state, mapping)
                children_ruleset.extend(child_ruleset)
                identicals.update(child_identicals)
            
            lhs_tree = Tree(self.label,
                    [str(i + 1) for i in range(len(child_states))])
#            print lhs_tree, child_states
            rhs = [("none", (s,)) for s in child_states]
#            print rhs
            r = Rule(
                ((state, lhs_tree),), self.label, rhs)
#            print r
            children_ruleset.append(r)
        return children_ruleset, identicals
#      except Exception as e:
#        print e
           
    def __init__(self, label, children):
        # a tree is a pair of label and list of children
        self.label = label
        self.children = children
    
    def frontier(self):
        # returns the yield
        if self.children:
            return " ".join([child.frontier() for child in self.children])
        else:
            return self.label
        
    def substitute(self, mapping):
        if self.label in mapping:
            return mapping[self.label] # item is an atomic tree
        else:
            return Tree(self.label, [child.substitute(mapping) for child in self.children])
        
    def __str__(self):
        if self.children:
            return "%s(%s)" % (self.label, " ".join(str(child) for child in self.children))
        else:
            return "%s" % (self.label,)
        
    def __eq__(self, other):
        if other is None:
            return False
        else:
            try:
                return self.label == other.label and all(
                    c == other.children[i] for (i, c) in enumerate(self.children))
            except(IndexError):
                return False
            except(KeyError):
                return False
            except(AttributeError):
                return False

    
    def compatible(self, other, is_var):
        if other is None:
            return False
        else:
            if is_var(self.label) or is_var(other.label):
                    return True
            else:
                try:
                    return self.label == other.label and all(
                        c.compatible(other.children[i], is_var) for (i, c) in enumerate(self.children))
                except(IndexError):
                    return False
                except(KeyError):
                    return False
                except(AttributeError):
                    return False
            
    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        try:
            return self._cached_hash
        except AttributeError:
            h = self._cached_hash = hash((self.label,
                tuple(hash(child) for child in self.children)))
            return h

    

class DagVertex:
    def __init__(self, label):
        self.label = label
        self.heads = []
        self.tails = []
        
    def __str__(self):
        return self.label
        
    def __repr__(self):
        return self.__str__()
        

class DagEdge:
    def __init__(self, head, tail, labels = []):
        self.head = head
        self.tail = tail
        self.labels = labels # list to hold states
    
    def __str__(self):
        return str((self.head, self.tail, self.labels))
     
    def __repr__(self):
        return self.__str__() 

          

         

class Dag:
    def __init__(self, from_file = None, from_tree = None, from_ve = None):
        def add_tree(tree, mother = None):
            vid = self.add_vertex(vlabel = tree.label)
            #print tree, mother, vid
            if mother is not None:
                self.add_edge(mother, vid)
            for child in tree.children:
                add_tree(child, vid) 
        
        if from_ve:
            self.V, self.E = copy.deepcopy(from_ve)
            self.num_vertices = len(self.V)
            self.root = self.V[0]
        else:            
            self.V = {}
            self.E = []
            self.num_vertices = 0
            if from_file:
                for line in from_file:
                    l = line.split()
                    if len(l) > 0:
                        if l[0] == "v":
                            self.add_vertex(*l[1:])
                        elif l[0] == "e":
                            self.add_edge(l[1], l[2], l[3:])
                        elif l[0] == "#":
                            print " ".join(l[1:])
            if from_tree:
                add_tree(from_tree)

    def __repr__(self):
        return self.__str__() 
    
    def add_pseudoroot(self):
        new_root = self.add_vertex()
        self.add_edge(new_root, self.root)
        self.root = new_root
        return new_root
        
    def remove_pseudoroot(self):
        del self.V[self.root]
        self.root = min(self.V.keys())
                  
    def add_vertex(self, vid = None, vlabel = None):
        if vid is None:
            vid = self.num_vertices
        if not self.V:
            self.root = vid
        self.V[vid] = DagVertex(vlabel)
        self.num_vertices += 1
        return vid
        
    def add_edge(self, head, tail, label = '', existing = False):
        new_edge = DagEdge(head, tail, label)
        edge_id = len(self.E)
        self.E.append(new_edge)
        if head not in self.V:
            self.add_vertex(head)
        if tail not in self.V:
            self.add_vertex(tail)
        if not existing:
            self.V[head].tails.append(edge_id)
            self.V[tail].heads.append(edge_id)
        return edge_id
        
    def topsort_vertices(self):
        return graphutils.topsort([(e.head, e.tail) for e in self.E])
    
    def dot(self, with_labels=True, bottomup=False):
        graph = pydot.Dot(graph_type='digraph', mode='hier',
            rankdir = 'BT' if bottomup else 'TB')
        nodes = {}
#        print self.V
        for (k,v) in self.V.iteritems(): 
            if with_labels:
                node = pydot.Node("%s (%s)" % (v.label, k))
            else:
                node = pydot.Node(str(k))
            nodes[k] = node
            node.set_shape('box' if '.' in str(k) else 'ellipse')
            graph.add_node(node)
        
#        print self.E
        for e in self.E:
            if e.tail != 0 and e.head != 0:
                graph.add_edge(pydot.Edge(nodes[e.head],
                    nodes[e.tail], label = str(e.labels) if e.labels else ''))
        return graph
                
    def __str__(self):
        #return "DAG with vertices %s and edges %s" % (self.V.keys(), self.E)
        return self.penman()
        
    def to_file(self, filename):
        if VERBOSE: print self.V
        if VERBOSE: print self.E
        f = open(filename, 'w')
        for vid in self.V:
            f.write("v %s %s\n" % (vid, self.V[vid].label))
        for vid in self.V:
            for tail in self.V[vid].tails:
                #print tail
                f.write("e %s %s\n" % (vid, self.E[tail].tail))
                
               
    def penman(self, v = None):
        traversed = set()
        
        def penman_e(e):
            try:
                return penman_v(self.E[e].tail)
            except TypeError:
                print "something wrong!", self.V, self.E
                return e
                
        
        def penman_v(vid):
            if vid in traversed:
                return str(vid)
            else:
                try:
                    sym = self.V[vid].label.split(":")
                    if sym == ['root']:
                        sym = ['root', '']
                    traversed.add(vid)
                    if sym[0] == "-":
                        return "-"
                    syms = [":%s" % (s, ) for s in sym[1:]]
                    return "(%s / %s)" % (vid, " ".join((sym[0],)
                        + tuple(" ".join(a) for a in zip(syms,
                        (penman_e(e) for e in self.V[vid].tails)))))
                except KeyError:
                    return "something wrong %s" % penman_v(self.root), self.V, self.E      
        return penman_v(self.root) if v is None else penman_v(v)
